package node

import (
	"bytes"
	"encoding/json"
	"log"
	"net"
	"os"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/autoradio"
	"git.autistici.org/ale/autoradio/instrumentation"
	"git.autistici.org/ale/autoradio/masterelection"
	"git.autistici.org/ale/autoradio/node/bwmonitor"
	"github.com/coreos/go-etcd/etcd"
)

var (
	icecastReloadErrors = instrumentation.NewCounter("icecast.reload_errors")
	icecastReloads      = instrumentation.NewCounter("icecast.reload")
	configIndex         = instrumentation.NewGauge("config.etcd_index")
)

func trigger(c chan bool) {
	select {
	case c <- true:
	default:
	}
}

// Converts an etcd key (a path) to the Icecast mount path. Removes
// mountPrefix from the beginning of the path, keeping the leading
// slash.
func keyToMount(key string) string {
	return key[len(autoradio.MountPrefix)-1:]
}

// In-memory representation of the overall configuration (basically
// just a list of the known mounts).
type clusterConfig struct {
	mounts map[string]*autoradio.Mount
	lock   sync.Mutex
}

func newClusterConfig() *clusterConfig {
	return &clusterConfig{
		mounts: make(map[string]*autoradio.Mount),
	}
}

// TODO: remove?
func (c *clusterConfig) GetMount(name string) *autoradio.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	return c.mounts[name]
}

func (c *clusterConfig) ListMounts() []*autoradio.Mount {
	c.lock.Lock()
	defer c.lock.Unlock()
	result := make([]*autoradio.Mount, 0, len(c.mounts))
	for _, m := range c.mounts {
		result = append(result, m)
	}
	return result
}

// Update a mount (in-memory only). Returns true if the new value is
// different than the old one.
func (c *clusterConfig) setMountIfChanged(m *autoradio.Mount) bool {
	c.lock.Lock()
	defer c.lock.Unlock()
	if prev, ok := c.mounts[m.Name]; ok && prev.Equal(m) {
		return false
	}
	c.mounts[m.Name] = m
	return true
}

// Delete a mount (in-memory only).
func (c *clusterConfig) delMount(name string) {
	c.lock.Lock()
	defer c.lock.Unlock()
	delete(c.mounts, name)
}

// Keeps the in-memory service configuration in sync with the etcd
// database. An update channel is triggered whenever the data changes.
type configWatcher struct {
	client autoradio.EtcdClient
	config *clusterConfig
	upch   chan bool
}

func newConfigSyncer(client autoradio.EtcdClient, config *clusterConfig, upch chan bool) *configWatcher {
	return &configWatcher{
		client: client,
		config: config,
		upch:   upch,
	}
}

// Returns true if the mount configuration has been modified.
func (w *configWatcher) updateMount(key, value string, index uint64) bool {
	mountName := keyToMount(key)
	var m autoradio.Mount
	if err := json.NewDecoder(strings.NewReader(value)).Decode(&m); err != nil {
		log.Printf("error updating mount %s [@%d]: corrupted data: %v", mountName, index, err)
	} else {
		if w.config.setMountIfChanged(&m) {
			log.Printf("updated mount %s [@%d]: %s", keyToMount(key), index, value)
			return true
		}
	}
	return false
}

func (w *configWatcher) watcher(index uint64, stop chan bool) {
	for {
		configIndex.Set(int64(index))
		resp, err := w.client.Watch(autoradio.MountPrefix, index+1, true, nil, stop)
		if err != nil {
			if err == etcd.ErrWatchStoppedByUser {
				return
			}
			if etcdErr, ok := err.(*etcd.EtcdError); ok {
				// If the index passed to Watch falls
				// behind the current log position by
				// more than 1000 entries (which is
				// bound to happen regularly in
				// presence of a sustained write
				// rate), etcd will return an error
				// 401 ("index out of date"). The safe
				// thing to do is just to run a full
				// load again, and resume watching
				// from the current EtcdIndex.
				if etcdErr.ErrorCode == 401 {
					index = w.load(stop)
					continue
				}
				// Log other etcd errors.
				log.Printf("config Watch(): %v", err)
				// Follow the etcd index even in case
				// of errors.
				index = etcdErr.Index
			} else {
				// Generic error, just retry the call.
				log.Printf("config Watch(): %v", err)
			}
			time.Sleep(200 * time.Millisecond)
			continue
		}

		index = resp.EtcdIndex

		switch resp.Action {
		case "delete":
			mountName := keyToMount(resp.Node.Key)
			log.Printf("deleted mount %s", mountName)
			w.config.delMount(mountName)
		case "set", "create", "update":
			if !w.updateMount(resp.Node.Key, resp.Node.Value, index) {
				// Do not trigger an update if the
				// update had no effect.
				continue
			}
		default:
			continue
		}

		// Trigger an update.
		trigger(w.upch)
	}
}

// Load full configuration from etcd. It will wait until the
// configuration can be read successfully, and it will trigger the
// update channel after it's done.
func (w *configWatcher) load(stop chan bool) uint64 {
	var err error
	var resp *etcd.Response
	for {
		if resp, err = w.client.Get(autoradio.MountPrefix, false, false); err != nil {
			log.Printf("config load error: %v", err)
		} else if resp.Node == nil || !resp.Node.Dir {
			log.Printf("config load error: no configuration found")
		} else {
			break
		}

		// Wait 1 second, but watch the stop channel.
		select {
		case <-time.After(1 * time.Second):
		case <-stop:
			return 0
		}
	}

	index := resp.EtcdIndex
	changed := false
	for _, n := range resp.Node.Nodes {
		if w.updateMount(n.Key, n.Value, index) {
			changed = true
		}
	}

	if changed {
		// Update the icecast daemon now that we have a full config.
		trigger(w.upch)
	}

	return index
}

// Start the configWatcher in the background. It will wait for
// initialization to complete, so that when this function returns, the
// in-memory configuration has already been fully synchronized.
func (w *configWatcher) Start(stop chan bool) func(chan bool) {
	log.Printf("loading configuration")
	index := w.load(stop)
	return func(stop chan bool) {
		w.watcher(index, stop)
	}
}

// An active streaming node, managing the local icecast server.
type RadioNode struct {
	client      autoradio.EtcdClient
	config      *clusterConfig
	name        string
	ips         []net.IP
	me          *masterelection.MasterElection
	watcher     *configWatcher
	icecast     Controller
	bw          *bwmonitor.BandwidthUsageMonitor
	heartbeat   uint64
	reloadDelay time.Duration
	upch        chan bool
	stop        chan bool
	wg          sync.WaitGroup
	Log         *log.Logger
}

func NewRadioNode(name string, ips []net.IP, netDev string, bwLimit float64, client autoradio.EtcdClient) *RadioNode {
	config := newClusterConfig()

	// Global 'stop' channel.
	stopch := make(chan bool)

	// Network updates trigger icecast reconfiguration. This
	// channel is used as an event signal.
	upch := make(chan bool, 1)

	// MasterElection changes trigger an update.
	mech := make(chan masterelection.State)
	go func() {
		for range mech {
			trigger(upch)
		}
	}()

	// Location information advertised when this node is master.
	minfo := &autoradio.MasterNodeInfo{
		Name: name,
		IP:   ips,
	}
	minfodata, err := json.Marshal(minfo)
	if err != nil {
		log.Fatal(err)
	}

	return &RadioNode{
		config: config,
		name:   name,
		ips:    ips,
		client: client,
		me: masterelection.NewMasterElection(
			client,
			autoradio.MasterElectionPath,
			string(minfodata),
			5,
			mech),
		watcher:     newConfigSyncer(client, config, upch),
		icecast:     NewIcecastController(name),
		reloadDelay: 1000 * time.Millisecond,
		heartbeat:   2,
		bw:          bwmonitor.NewBandwidthUsageMonitor(netDev, bwLimit),
		upch:        upch,
		stop:        stopch,
		Log:         log.New(os.Stderr, "node: ", 0),
	}
}

// The presence goroutine periodically updates our entry in the list
// of nodes with the current node statistics.
func (rc *RadioNode) presence(stop chan bool) {
	ticker := time.NewTicker(time.Duration(rc.heartbeat) * time.Second / 3)

	// Register ourselves using the node name.
	key := autoradio.NodePrefix + rc.name

	for {
		select {
		case <-ticker.C:
			// Build our NodeStatus.
			icecastStatus := rc.icecast.GetStatus()
			nodeStatus := autoradio.NodeStatus{
				Name:           rc.name,
				IP:             rc.ips,
				IcecastUp:      icecastStatus.Up,
				Mounts:         icecastStatus.Mounts,
				BandwidthUsage: rc.bw.GetUsage(),
			}

			// Update our node entry in the database.
			var buf bytes.Buffer
			json.NewEncoder(&buf).Encode(&nodeStatus)
			if _, err := rc.client.Set(key, buf.String(), rc.heartbeat); err != nil {
				rc.Log.Printf("presence: Set(): %s", err)
			}

		case <-stop:
			return
		}
	}
}

// Get a valid IP address for the current master node (to be passed to
// Icecast). Since we don't really know much about network topology,
// just pick the first IP address associated with the master node.
func (rc *RadioNode) getMasterAddr() net.IP {
	var info autoradio.MasterNodeInfo
	if err := json.NewDecoder(strings.NewReader(rc.me.GetMasterData())).Decode(&info); err != nil || len(info.IP) == 0 {
		return nil
	}
	return info.IP[0]
}

// Reload the icecast configuration when needed.
func (rc *RadioNode) updater(stop chan bool) {
	// Wait an instant to give a chance to the services to
	// initialize properly.
	time.Sleep(200 * time.Millisecond)

	rc.Log.Printf("starting icecast updater")
	for {
		select {
		case <-rc.upch:
			// We may have received an update before the
			// masterelection had time to actually elect a
			// master. In this case, skip.
			if !rc.me.Valid() {
				continue
			}

			icecastReloads.Incr()
			rc.Log.Printf("reloading icecast config")
			if err := rc.icecast.Update(rc.config, rc.me.IsMaster(), rc.getMasterAddr()); err != nil {
				icecastReloadErrors.Incr()
				rc.Log.Printf("Update(): %s", err)
			}

			// Limit the rate of icecast reloads.
			if rc.reloadDelay > 0 {
				time.Sleep(rc.reloadDelay)
			}

		case <-stop:
			return
		}
	}
}

// Start the node. Returns immediately.
func (rc *RadioNode) Start() {
	// Bootstrap the config watcher first. This ensures that we
	// have a full configuration (thanks to the Get() call) before
	// we start managing the icecast server.
	watcherFn := rc.watcher.Start(rc.stop)

	// Start auxiliary services and event listeners on their own
	// goroutines. Put them all in a WaitGroup so we can wait for
	// their termination.
	bgfuncs := []func(chan bool){
		// Config watcher.
		watcherFn,

		// Presence heartbeat.
		rc.presence,

		// Icecast updater,
		rc.updater,

		// Master election runner.
		rc.me.Run,

		// Icecast status checker.
		rc.icecast.Run,

		// Bandwidth monitor.
		rc.bw.Run,
	}

	for _, fn := range bgfuncs {
		rc.wg.Add(1)
		go func(fn func(stop chan bool)) {
			fn(rc.stop)
			rc.wg.Done()
		}(fn)
	}
}

// Wait until all processing associated with this node has terminated.
func (rc *RadioNode) Wait() {
	rc.wg.Wait()
	// Let's leave this around for garbage collection, otherwise
	// the in-memory etcd mock server might try to send values to
	// it due to a Watch-trigger desynchronization...
	//close(rc.upch)
}

// Run the node, waiting for termination.
func (rc *RadioNode) Run() {
	rc.Start()
	rc.Wait()
}

// Stop everything.
func (rc *RadioNode) Stop() {
	close(rc.stop)
}
